# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import DimensionClass
from .....schema_classes import ModelDatasetClass
from .....schema_classes import SemanticFieldClass
from .....schema_classes import SemanticFieldTypeClass
from .....schema_classes import SemanticModelInfoClass
from .....schema_classes import SemanticModelRelationshipClass


Dimension = DimensionClass
ModelDataset = ModelDatasetClass
SemanticField = SemanticFieldClass
SemanticFieldType = SemanticFieldTypeClass
SemanticModelInfo = SemanticModelInfoClass
SemanticModelRelationship = SemanticModelRelationshipClass

# fmt: on
