# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import AiContextClass
from .....schema_classes import DerivedMetricInputClass
from .....schema_classes import DialectClass
from .....schema_classes import DialectExpressionClass
from .....schema_classes import MetricExpressionClass
from .....schema_classes import MetricInfoClass
from .....schema_classes import MetricRelationshipsClass
from .....schema_classes import MetricUpstreamsClass


AiContext = AiContextClass
DerivedMetricInput = DerivedMetricInputClass
Dialect = DialectClass
DialectExpression = DialectExpressionClass
MetricExpression = MetricExpressionClass
MetricInfo = MetricInfoClass
MetricRelationships = MetricRelationshipsClass
MetricUpstreams = MetricUpstreamsClass

# fmt: on
